"""
There's no such thing as an "S3 grid job". The code in this module just helps us
implement grid jobs that use an S3-compatible object store to transfer data, e.g. for
EC2 and Kubernetes
"""


from __future__ import annotations

import asyncio
import io
import pickle
import time
import typing

from typing import (
    Any,
    AsyncIterable,
    Iterable,
    List,
    Optional,
    Set,
    TYPE_CHECKING,
    Tuple,
)

import aiobotocore.session
import botocore.exceptions
from meadowrun.aws_integration import s3

from meadowrun.meadowrun_pb2 import ProcessState
from meadowrun.run_job_core import (
    JobCompletion,
    MeadowrunException,
    TaskProcessState,
    WorkerProcessState,
)
from meadowrun.storage_keys import (
    STORAGE_KEY_PROCESS_STATE_SUFFIX,
    STORAGE_KEY_TASK_RESULT_SUFFIX,
    parse_storage_key_process_state,
    parse_storage_key_task_result,
    storage_key_process_state,
    storage_key_task_args,
    storage_key_task_result,
    storage_prefix_outputs,
)

if TYPE_CHECKING:
    import types_aiobotocore_s3
    from typing_extensions import Literal


def get_storage_client_from_args(
    storage_endpoint_url: Optional[str],
    storage_access_key_id: Optional[str],
    storage_secret_access_key: Optional[str],
) -> typing.AsyncContextManager[types_aiobotocore_s3.S3Client]:
    kwargs = {}
    if storage_access_key_id is not None:
        kwargs["aws_access_key_id"] = storage_access_key_id
    if storage_secret_access_key is not None:
        kwargs["aws_secret_access_key"] = storage_secret_access_key
    if storage_endpoint_url is not None:
        kwargs["endpoint_url"] = storage_endpoint_url

    # TODO if all the parameters are None then we're implicitly falling back on AWS
    # S3, which we should make explicit
    session = aiobotocore.session.get_session()
    return session.create_client("s3", **kwargs)  # type: ignore


async def read_storage(
    storage_client: types_aiobotocore_s3.S3Client,
    storage_bucket: str,
    storage_filename: str,
) -> Any:
    response = await storage_client.get_object(
        Bucket=storage_bucket, Key=storage_filename
    )
    async with response["Body"] as stream:
        return await stream.read()


async def write_storage_pickle(
    storage_client: types_aiobotocore_s3.S3Client,
    storage_bucket: str,
    storage_filename: str,
    obj: Any,
    pickle_protocol: Optional[int],
) -> None:
    with io.BytesIO() as buffer:
        pickle.dump(obj, buffer, protocol=pickle_protocol)
        buffer.seek(0)
        await storage_client.put_object(
            Bucket=storage_bucket,
            Key=storage_filename,
            Body=buffer,
        )


async def write_storage_file(
    storage_client: types_aiobotocore_s3.S3Client,
    storage_bucket: str,
    local_filename: str,
    storage_filename: str,
) -> None:
    with open(local_filename, "rb") as f:
        await storage_client.put_object(
            Bucket=storage_bucket, Key=storage_filename, Body=f
        )


async def try_get_storage_file(
    storage_client: types_aiobotocore_s3.S3Client,
    storage_bucket: str,
    storage_filename: str,
    local_filename: str,
) -> bool:
    try:
        response = await storage_client.get_object(
            Bucket=storage_bucket, Key=storage_filename
        )
        async with response["Body"] as stream:
            with open(local_filename, "wb") as f:
                f.write(await stream.read())
        return True
    except botocore.exceptions.ClientError as error:
        # don't raise an error saying the file doesn't exist, we'll just upload it
        # in that case by falling through to the next bit of code
        if error.response["Error"]["Code"] not in ("404", "NoSuchKey"):
            raise error

        return False


async def upload_task_args(
    s3_client: types_aiobotocore_s3.S3Client,
    bucket_name: str,
    job_id: str,
    args: Iterable[Any],
) -> List[Tuple[int, int]]:
    range_from = 0
    ranges = []
    with io.BytesIO() as buffer:
        for i, arg in enumerate(args):
            pickle.dump(((arg,), {}), buffer)
            range_to = buffer.tell() - 1
            ranges.append((range_from, range_to))
            range_from = range_to + 1

        await s3_client.put_object(
            Bucket=bucket_name,
            Key=storage_key_task_args(job_id),
            Body=buffer.getvalue(),
        )

    return ranges


async def download_task_arg(
    s3_client: types_aiobotocore_s3.S3Client,
    bucket_name: str,
    job_id: str,
    byte_range: Tuple[int, int],
) -> Any:
    return (
        await s3.download_async(
            s3_client, bucket_name, storage_key_task_args(job_id), byte_range
        )
    )[1]


async def complete_task(
    s3_client: types_aiobotocore_s3.S3Client,
    bucket_name: str,
    job_id: str,
    task_id: int,
    attempt: int,
    process_state: ProcessState,
) -> None:
    """Uploads the result of the task to S3."""
    await s3_client.put_object(
        Bucket=bucket_name,
        Key=storage_key_task_result(job_id, task_id, attempt),
        Body=process_state.SerializeToString(),
    )


async def receive_results(
    s3_client: types_aiobotocore_s3.S3Client,
    bucket_name: str,
    job_id: str,
    stop_receiving: asyncio.Event,
    all_workers_exited: asyncio.Event,
    initial_wait_seconds: int = 1,
    receive_message_wait_seconds: int = 20,
    read_worker_process_states: bool = True,
) -> AsyncIterable[Tuple[List[TaskProcessState], List[WorkerProcessState]]]:
    """
    Listens to a result queue until we have results for num_tasks.

    As results become available, yields (task results, worker results). Task results
    will be a list of TaskProcessState. Worker results will be a list of
    WorkerProcessState
    """

    # Behavior is that if stop_receiving is set, we want to return immediately. If
    # all_workers_exited is set, then keep trying for about 3 seconds (just in case some
    # results are still coming in), and then return

    delete_tasks = []

    try:
        results_prefix = storage_prefix_outputs(job_id)
        download_keys_received: Set[str] = set()
        wait = initial_wait_seconds
        workers_exited_wait_count = 0
        while not stop_receiving.is_set() and (
            workers_exited_wait_count < 3 or wait == 0
        ):
            if all_workers_exited.is_set():
                workers_exited_wait_count += 1

            if wait:
                events_to_wait_for = [asyncio.create_task(stop_receiving.wait())]
                if workers_exited_wait_count == 0:
                    events_to_wait_for.append(
                        asyncio.create_task(all_workers_exited.wait())
                    )
                else:
                    # poll more frequently if workers are done, but still wait 1 second
                    # (unless stop_receiving is set)
                    wait = 1

                done, pending = await asyncio.wait(
                    events_to_wait_for,
                    timeout=wait,
                    return_when=asyncio.FIRST_COMPLETED,
                )
                for p in pending:
                    p.cancel()
                if stop_receiving.is_set():
                    break

            keys = await s3.list_objects_async(
                s3_client, bucket_name, results_prefix, ""
            )

            download_tasks = []
            for key in keys:
                if key not in download_keys_received:
                    download_tasks.append(
                        asyncio.create_task(
                            s3.download_async(s3_client, bucket_name, key)
                        )
                    )

            download_keys_received.update(keys)

            if len(download_tasks) == 0:
                if wait == 0:
                    wait = 1
                else:
                    wait = min(wait * 2, receive_message_wait_seconds)
            else:
                wait = 0
                task_results = []
                worker_results = []
                for task_result_future in asyncio.as_completed(download_tasks):
                    key, process_state_bytes = await task_result_future
                    process_state = ProcessState()
                    process_state.ParseFromString(process_state_bytes)
                    if key.endswith(STORAGE_KEY_TASK_RESULT_SUFFIX):
                        task_id, attempt = parse_storage_key_task_result(
                            key, results_prefix
                        )
                        task_results.append(
                            TaskProcessState(task_id, attempt, process_state)
                        )
                    elif key.endswith(STORAGE_KEY_PROCESS_STATE_SUFFIX):
                        if not read_worker_process_states:
                            continue

                        worker_index = parse_storage_key_process_state(
                            key, results_prefix
                        )
                        worker_results.append(
                            WorkerProcessState(worker_index, process_state)
                        )
                    else:
                        print(f"Warning, unrecognized key {key}, will ignore")
                        continue  # don't delete if we can't parse the key

                    delete_tasks.append(
                        asyncio.create_task(
                            s3_client.delete_object(Bucket=bucket_name, Key=key)
                        )
                    )
                yield task_results, worker_results
    finally:
        await asyncio.gather(*delete_tasks, return_exceptions=True)


async def get_job_completion_from_process_state(
    storage_client: Any,
    storage_bucket: str,
    job_id: str,
    worker_index: str,
    job_spec_type: Literal["py_command", "py_function", "py_agent"],
    timeout_seconds: int,
    public_address: str,
) -> JobCompletion[Any]:
    """
    Polls for the specified .process_state file, and then creates a JobCompletion object
    based on the process_state

    file_suffix is used by indexed completion jobs to distinguish between the
    completions of different workers, this should be set to the job completion index.
    """

    t0 = time.time()
    process_state_bytes = None
    process_state_key = storage_key_process_state(job_id, worker_index)
    wait = 1
    while time.time() - t0 < timeout_seconds:
        try:
            process_state_bytes = await read_storage(
                storage_client,
                storage_bucket,
                process_state_key,
            )
            break
        except botocore.exceptions.ClientError as error:
            if error.response["Error"]["Code"] not in ("404", "NoSuchKey"):
                raise

        await asyncio.sleep(wait)
        wait = max(wait + 1, 20)

    if process_state_bytes is None:
        raise TimeoutError(
            f"Waited {timeout_seconds} seconds but {process_state_key} does not exist"
        )

    process_state = ProcessState.FromString(process_state_bytes)
    if process_state.state == ProcessState.ProcessStateEnum.SUCCEEDED:
        # we must have a result from functions, in other cases we can optionally have a
        # result
        if job_spec_type == "py_function" or process_state.pickled_result:
            result = pickle.loads(process_state.pickled_result)
        else:
            result = None
        return JobCompletion(
            result,
            process_state.state,
            process_state.log_file_name,
            process_state.return_code,
            public_address,
        )
    else:
        raise MeadowrunException(process_state)



from __future__ import annotations

import abc
import os
import shutil
import sys
import urllib.parse
import zipfile
from typing import TYPE_CHECKING, Tuple, Type

import filelock

if TYPE_CHECKING:
    from types import TracebackType


class ObjectStorage(abc.ABC):
    """An ObjectStorage is a place where you can upload files and download them"""

    async def __aenter__(self) -> ObjectStorage:
        return self

    async def __aexit__(
        self,
        exc_typ: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        pass

    @classmethod
    @abc.abstractmethod
    def get_url_scheme(cls) -> str:
        """
        Right now we're using the URL scheme to effectively serialize the ObjectStorage
        object to the job. This works as long as we don't need any additional parameters
        (region name, username/password), but we may need to make this more flexible in
        the future.
        """
        pass

    async def upload_from_file_url(self, file_url: str) -> str:
        """
        file_url will be a file:// url to a file on the local machine. This function
        should upload that file to the object storage, delete the local file, and return
        the URL of the remote file.
        """
        decoded_url = urllib.parse.urlparse(file_url)
        if decoded_url.scheme != "file":
            raise ValueError(f"Expected file URI: {file_url}")
        if sys.platform == "win32" and decoded_url.path.startswith("/"):
            # on Windows, file:///C:\foo turns into file_url.path = /C:\foo so we need
            # to remove the forward slash at the beginning
            file_path = decoded_url.path[1:]
        else:
            file_path = decoded_url.path
        bucket_name, object_name = await self._upload(file_path)
        shutil.rmtree(os.path.dirname(file_path), ignore_errors=True)
        return urllib.parse.urlunparse(
            (self.get_url_scheme(), bucket_name, object_name, "", "", "")
        )

    async def download_and_unzip(
        self, remote_url: str, local_copies_folder: str
    ) -> str:
        """
        remote_url will be the URL of a file in the object storage system as generated
        by upload_from_file_url. This function should download the file and extract it
        to local_copies_folder if it has not already been extracted.
        """
        decoded_url = urllib.parse.urlparse(remote_url)
        bucket_name = decoded_url.netloc
        object_name = decoded_url.path.lstrip("/")
        extracted_folder = os.path.join(
            local_copies_folder, os.path.basename(object_name)
        )

        with filelock.FileLock(f"{extracted_folder}.lock", timeout=120):
            if not os.path.exists(extracted_folder):
                zip_file_path = extracted_folder + ".zip"
                await self._download(bucket_name, object_name, zip_file_path)
                with zipfile.ZipFile(zip_file_path) as zip_file:
                    zip_file.extractall(extracted_folder)

        return extracted_folder

    @abc.abstractmethod
    async def _upload(self, file_path: str) -> Tuple[str, str]:
        """Ensure the file at the given file path is uploaded to storage. Returns where
        the file was uploaded as a tuple of bucket name and object name."""
        pass

    @abc.abstractmethod
    async def _download(
        self, bucket_name: str, object_name: str, file_name: str
    ) -> None:
        """Download the file at given bucket and object name locally to the given file
        name."""
        pass



from __future__ import annotations

import hashlib
import os
from typing import Tuple

from meadowrun.azure_integration.azure_meadowrun_core import (
    ensure_meadowrun_storage_account,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_exceptions import (
    AzureRestApiError,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_rest_api import (
    azure_rest_api,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_storage_api import (
    StorageAccount,
    azure_blob_api,
)
from meadowrun.azure_integration.mgmt_functions.azure_constants import (
    MEADOWRUN_RESOURCE_GROUP_NAME,
)
from meadowrun.object_storage import ObjectStorage

CONTAINER_NAME = "meadowrun"


_LIFECYCLE_POLICY = {
    "rules": [
        {
            "enabled": True,
            "name": "meadowrun-lifecycle-policy",
            "type": "Lifecycle",
            "definition": {
                "actions": {
                    "baseBlob": {"delete": {"daysAfterModificationGreaterThan": 14}}
                },
                "filters": {"blobTypes": ["blockBlob"], "prefixMatch": ["meadowrun/"]},
            },
        }
    ]
}


async def ensure_container(
    subscription_id: str, storage_account: StorageAccount, location: str
) -> None:
    storage_account = await ensure_meadowrun_storage_account(location, "create")
    # https://docs.microsoft.com/en-us/rest/api/storageservices/create-container
    await azure_blob_api(
        "PUT",
        storage_account,
        CONTAINER_NAME,
        query_parameters={"restype": "container"},
        ignored_status_codes=((409, "ContainerAlreadyExists"),),
    )

    await azure_rest_api(
        "PUT",
        f"subscriptions/{subscription_id}/resourceGroups/"
        f"{MEADOWRUN_RESOURCE_GROUP_NAME}/providers/Microsoft.Storage/storageAccounts/"
        f"{storage_account.name}/managementPolicies/default",
        api_version="2021-09-01",
        json_content={"properties": {"policy": _LIFECYCLE_POLICY}},
    )


async def ensure_uploaded(file_path: str) -> Tuple[str, str]:
    hasher = hashlib.blake2b()
    with open(file_path, "rb") as file:
        buf = file.read()
        hasher.update(buf)
    digest = hasher.hexdigest()

    # the empty location is inelegant here, but:
    # 1. it's only used if the storage account does not exist yet; in that case we're
    #    not even getting to this point
    # 2. it's pretty annoying to thread it where download_file is used.
    storage_account = await ensure_meadowrun_storage_account("", on_missing="raise")
    try:
        _ = await azure_blob_api("HEAD", storage_account, f"{CONTAINER_NAME}/{digest}")
        # CONTAINER_NAME is a constant on Azure, but not AWS. Returning it here for
        # consistency.
        return CONTAINER_NAME, digest
    except AzureRestApiError as error:
        if error.status != 404:
            raise error

    # doesn't exist, need to upload it
    content_length = os.path.getsize(file_path)
    with open(file_path, "rb") as file:
        _ = await azure_blob_api(
            "PUT",
            storage_account,
            f"{CONTAINER_NAME}/{digest}",
            additional_headers={
                "Content-Length": str(content_length),
                "x-ms-blob-type": "BlockBlob",
            },
            binary_content=file,
        )
    return CONTAINER_NAME, digest


async def download_file(container_name: str, object_name: str, file_name: str) -> None:
    # see comment in ensure_uploaded_by_hash regarding empty location
    storage_account = await ensure_meadowrun_storage_account("", on_missing="raise")

    result: bytes = await azure_blob_api(
        "GET", storage_account, f"{container_name}/{object_name}"
    )
    with open(file_name, "wb") as file:
        file.write(result)


class AzureBlobStorage(ObjectStorage):
    # TODO this should have a region_name property also

    @classmethod
    def get_url_scheme(cls) -> str:
        return "azblob"

    async def _upload(self, file_path: str) -> Tuple[str, str]:
        return await ensure_uploaded(file_path)

    async def _download(
        self, bucket_name: str, object_name: str, file_name: str
    ) -> None:
        return await download_file(bucket_name, object_name, file_name)



from __future__ import annotations

import dataclasses
import hashlib
from io import BytesIO
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

import aiobotocore.session
import boto3
import boto3.exceptions
from botocore.exceptions import ClientError

from meadowrun.aws_integration.aws_core import (
    MeadowrunAWSAccessError,
    MeadowrunNotInstalledError,
    _get_account_number,
    _get_default_region_name,
)
from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
    ignore_boto3_error_code,
)
from meadowrun.object_storage import ObjectStorage
from meadowrun.storage_keys import STORAGE_CODE_CACHE_PREFIX

if TYPE_CHECKING:
    from types_aiobotocore_s3.client import S3Client

BUCKET_PREFIX = "meadowrun"


def _get_bucket_name(region_name: str) -> str:
    # s3 bucket names must be globally unique across all accounts and regions.
    return f"{BUCKET_PREFIX}-{region_name}-{_get_account_number()}"


def ensure_bucket(
    region_name: str,
    expire_days: int = 14,
) -> None:
    """
    Create an S3 bucket in a specified region if it does not exist yet.

    If a region is not specified, the bucket is created in the configured default
    region.

    The bucket is created with a default lifecycle policy of 14 days.

    Since bucket names must be globally unique, the name is bucket_prefix + region +
    account number

    :param bucket_name: Bucket to create
    :param region_name: String region to create bucket in, e.g., 'us-west-2'
    :param expire_days: int number of days after which keys are deleted by lifecycle
        policy.
    :return: the full bucket name
    """

    s3 = boto3.client("s3", region_name=region_name)

    bucket_name = _get_bucket_name(region_name)

    # us-east-1 cannot be specified as a LocationConstraint because it is the default
    # region
    # https://stackoverflow.com/questions/51912072/invalidlocationconstraint-error-while-creating-s3-bucket-when-the-used-command-i
    if region_name == "us-east-1":
        additional_parameters = {}
    else:
        additional_parameters = {
            "CreateBucketConfiguration": {"LocationConstraint": region_name}
        }

    success, _ = ignore_boto3_error_code(
        lambda: s3.create_bucket(Bucket=bucket_name, **additional_parameters),
        "BucketAlreadyOwnedByYou",
    )
    s3.put_bucket_lifecycle_configuration(
        Bucket=bucket_name,
        LifecycleConfiguration=dict(
            Rules=[
                dict(
                    Expiration=dict(
                        Days=expire_days,
                    ),
                    ID="meadowrun-lifecycle-policy",
                    # Filter is mandatory, but we don't want one:
                    Filter=dict(Prefix=""),
                    Status="Enabled",
                )
            ]
        ),
    )


async def ensure_uploaded_by_hash(
    s3_client: S3Client,
    local_file_path: str,
    bucket_name: str,
    key_prefix: str = "",
) -> str:
    """
    Uploads the specified file from the local machine to S3. The file will be uploaded
    to the key "{key_prefix}{hash}", where {hash} is a hash of the contents of the file.
    If the S3 object already exists, it does not need to be re-uploaded.

    key_prefix should usually be "" or end in a "/" like "code/". key_suffix should
    usually be "" or something like ".tar.gz"

    Returns the key of the file in the S3 bucket, e.g. "code/123456789abcdefg"
    """

    hasher = hashlib.blake2b()
    with open(local_file_path, "rb") as file:
        buf = file.read()
        hasher.update(buf)
    s3_key = f"{key_prefix}{hasher.hexdigest()}"

    try:
        await s3_client.head_object(Bucket=bucket_name, Key=s3_key)
        return s3_key
    except ClientError as error:
        if error.response["Error"]["Code"] == "403":
            # if we don't have permissions to the bucket, throw a helpful error
            raise MeadowrunAWSAccessError("S3 bucket") from error

        # don't raise an error saying the file doesn't exist, we'll just upload it in
        # that case by falling through to the next bit of code
        if not error.response["Error"]["Code"] == "404":
            raise error

    # doesn't exist, need to upload it
    try:
        await s3_client.put_object(Bucket=bucket_name, Key=s3_key, Body=buf)
    except boto3.exceptions.S3UploadFailedError as e:
        if len(e.args) >= 1 and "NoSuchBucket" in e.args[0]:
            raise MeadowrunNotInstalledError("S3 bucket")
        raise

    return s3_key


async def download_file(
    bucket_name: str,
    object_name: str,
    file_name: str,
    region_name: Optional[str] = None,
) -> None:
    if region_name is None:
        region_name = await _get_default_region_name()

    s3 = boto3.client("s3", region_name=region_name)

    s3.download_file(bucket_name, object_name, file_name)


def upload(object_name: str, data: bytes, region_name: str) -> None:
    s3 = boto3.client("s3", region_name=region_name)
    bucket_name = _get_bucket_name(region_name)
    with BytesIO(data) as file_obj:
        s3.upload_fileobj(file_obj, bucket_name, object_name)


async def upload_async(
    object_name: str, data: bytes, region_name: str, s3_client: Any
) -> None:
    bucket_name = _get_bucket_name(region_name)
    await s3_client.put_object(Bucket=bucket_name, Key=object_name, Body=data)


async def list_objects_async(
    s3_client: Any,
    bucket_name: str,
    prefix: str,
    start_after: str,
) -> List[str]:
    """Returns the keys in the meadowrun bucket."""
    paginator = s3_client.get_paginator("list_objects_v2")
    results = []
    async for result in paginator.paginate(
        Bucket=bucket_name, Prefix=prefix, StartAfter=start_after
    ):
        for c in result.get("Contents", []):
            results.append(c["Key"])
    return results


async def download_async(
    s3c: S3Client,
    bucket_name: str,
    object_name: str,
    byte_range: Optional[Tuple[int, int]] = None,
) -> Tuple[str, bytes]:
    if byte_range is None:
        response = await s3c.get_object(Bucket=bucket_name, Key=object_name)
    else:
        response = await s3c.get_object(
            Bucket=bucket_name,
            Key=object_name,
            Range=f"bytes={byte_range[0]}-{byte_range[1]}",
        )
    async with response["Body"] as stream:
        return object_name, await stream.read()


def delete_bucket(region_name: str) -> None:
    """Deletes the meadowrun bucket"""
    s3 = boto3.resource("s3", region_name=region_name)

    bucket = s3.Bucket(_get_bucket_name(region_name))
    success, _ = ignore_boto3_error_code(
        lambda: list(bucket.objects.limit(1)),
        "NoSuchBucket",
    )
    if success:
        # S3 doesn't allow deleting a bucket with anything in it, so delete all objects
        # in chunks of up to 1000, which is the maximum allowed.
        key_chunk: List[Dict[str, str]] = []
        for s3object in bucket.objects.all():
            if len(key_chunk) == 1000:
                bucket.delete_objects(Delete=dict(Objects=key_chunk))
                key_chunk.clear()
            key_chunk.append(dict(Key=s3object.key))
        if key_chunk:
            bucket.delete_objects(Delete=dict(Objects=key_chunk))

        bucket.delete()


@dataclasses.dataclass
class S3ObjectStorage(ObjectStorage):
    region_name: Optional[str] = None

    @classmethod
    def get_url_scheme(cls) -> str:
        return "s3"

    async def _upload(self, file_path: str) -> Tuple[str, str]:
        if self.region_name is None:
            region_name = await _get_default_region_name()
        else:
            region_name = self.region_name

        bucket_name = _get_bucket_name(region_name)

        async with aiobotocore.session.get_session().create_client(
            "s3", region_name=region_name
        ) as s3_client:
            try:
                s3_key = await ensure_uploaded_by_hash(
                    s3_client, file_path, bucket_name, STORAGE_CODE_CACHE_PREFIX
                )
            except ClientError as error:
                if error.response["Error"]["Code"] == "403":
                    # if we don't have permissions to the bucket, throw a helpful error
                    raise MeadowrunAWSAccessError("S3 bucket") from error

                raise

        return bucket_name, s3_key

    async def _download(
        self, bucket_name: str, object_name: str, file_name: str
    ) -> None:
        return await download_file(
            bucket_name, object_name, file_name, self.region_name
        )



"""
This module contains functionality for interacting with Kubernetes that theoretically
has nothing to do with Meadowrun. I.e. you can think of this as a slightly higher-level
version of the kubernetes_asyncio module.
"""

from __future__ import annotations

import asyncio
import base64
import time
from typing import Callable, List, Dict, Tuple, Optional

from kubernetes_asyncio import client as kubernetes_client, watch as kubernetes_watch

from meadowrun.run_job_core import WaitOption


# kubernetes_asyncio.watch.Watch looks at the docstring to figure out if the function
# passed to Watch takes a follow or watch argument. This results in the wrong value for
# read_namespaced_pod_log, so we just monkey-patch the function here
_orig_get_watch_argument_name = kubernetes_watch.Watch.get_watch_argument_name


def _new_get_watch_argument_name(watch: kubernetes_watch.Watch, func: Callable) -> str:
    if getattr(func, "__name__") == "read_namespaced_pod_log":
        return "follow"
    return _orig_get_watch_argument_name(watch, func)


kubernetes_watch.watch.Watch.get_watch_argument_name = _new_get_watch_argument_name


async def get_pods_for_job(
    core_api: kubernetes_client.CoreV1Api,
    kubernetes_namespace: str,
    job_id: str,
    pod_generate_names: List[str],
) -> List[kubernetes_client.V1Pod]:
    """
    When you launch a Kubernetes job, one or more pods get created. In our case, we
    should only ever get one pod for regular jobs, and one pod for each index in indexed
    completion jobs, because we always set parallelism = completions, and we've
    configured 0 retries.

    For regular jobs, the pod created will be named <job_id>-<random string>. The pod's
    metadata has a generate_name, which will be equal to <job_id>-.

    For indexed completion jobs, there will be a pod for each index with a generate_name
    of <job_id>-<index>-.

    pod_generate_names should have the list of generate_names that we expect to see for
    the specified job.

    Returns a pod object corresponding to each pod_generate_names, in the same order as
    pod_generate_names.
    """

    pod_generate_names_set = set(pod_generate_names)
    results: Dict[str, kubernetes_client.V1Pod] = {}
    i = 0

    while True:
        pods = await core_api.list_namespaced_pod(
            kubernetes_namespace, label_selector=f"job-name={job_id}"
        )
        for pod in pods.items:
            generate_name = pod.metadata.generate_name

            if generate_name not in pod_generate_names_set:
                raise ValueError(
                    f"Unexpected pod {pod.metadata.name} with generate name "
                    f"{generate_name} found"
                )

            if (
                generate_name in results
                and results[generate_name].metadata.name != pod.metadata.name
            ):
                # TODO we may need to change this if we add e.g. retries
                raise ValueError(
                    "Unexpected multiple pods with the same generate name "
                    f"{generate_name} found: {results[generate_name].metadata.name}, "
                    f"{pod.metadata.name}"
                )

            results[generate_name] = pod

        if len(results) >= len(pod_generate_names):
            break

        if i > 15:
            raise TimeoutError(
                "Waited >15s, but pods with the following generate names were not "
                "created: "
                + ", ".join(p for p in pod_generate_names if p not in results)
            )

        if i == 0:
            print(f"Waiting for pods to be created for the job {job_id}")

        await asyncio.sleep(1.0)

        i += 1

    return [results[generate_name] for generate_name in pod_generate_names]


def get_main_container_state(
    pod: kubernetes_client.V1Pod, job_id: str, pod_name: str
) -> Tuple[Optional[kubernetes_client.V1ContainerState], Optional[str]]:
    # first get main container state
    container_statuses = pod.status.container_statuses

    if container_statuses is None or len(container_statuses) == 0:
        main_container_state = None
    else:
        main_container_statuses = [s for s in container_statuses if s.name == "main"]
        if len(main_container_statuses) == 0:
            raise ValueError(
                f"The job {job_id} has a pod {pod_name} but there is no `main` "
                "container"
            )
        if len(main_container_statuses) > 1:
            raise ValueError(
                f"The job {job_id} has a pod {pod_name} but there is more than one "
                "`main` container"
            )
        main_container_state = main_container_statuses[0].state

    # then get the latest condition's reason, this is where Kubernetes will tell us that
    # e.g. the pod is unschedulable
    if pod.status.conditions:
        latest_condition = pod.status.conditions[-1]
        result_builder = []
        if latest_condition.reason:
            result_builder.append(latest_condition.reason)
        if latest_condition.message:
            result_builder.append(latest_condition.message)
        latest_condition_reason = ", ".join(result_builder)
    else:
        latest_condition_reason = None

    return main_container_state, latest_condition_reason


async def wait_for_pod_running(
    core_api: kubernetes_client.CoreV1Api,
    job_id: str,
    kubernetes_namespace: str,
    pod: kubernetes_client.V1Pod,
) -> kubernetes_client.V1Pod:
    pod_name = pod.metadata.name

    # The first step is to wait for the pod to start running, because we can't stream
    # logs until the pod is in a running state. The happy path is that our pod is in the
    # "waiting" state because we're either waiting for the image to get pulled or we're
    # waiting for nodes to become available to run our job. In that case, we'll wait up
    # to 7 minutes.
    #
    # The unhappy path is that something has gone wrong which Kubernetes expresses as
    # waiting infinitely, rather than a failure. E.g. if our image spec is invalid. In
    # that case we'll only wait 15 seconds, as it doesn't make sense to expect that that
    # would change.

    i = 0
    wait_until = 15
    max_wait_until = 60 * 7
    main_container_state, latest_condition_reason = get_main_container_state(
        pod, job_id, pod_name
    )
    prev_additional_info = None
    while main_container_state is None or (
        main_container_state.running is None and main_container_state.terminated is None
    ):
        is_happy_path = False
        additional_info_builder = [":"]
        if (
            main_container_state is not None
            and main_container_state.waiting is not None
        ):
            additional_info_builder.append(str(main_container_state.waiting.reason))
            if main_container_state.waiting.message is not None:
                additional_info_builder.append(
                    str(main_container_state.waiting.message)
                )
            elif main_container_state.waiting.reason == "ContainerCreating":
                # TODO Kubernetes unfortunately doesn't distinguish between waiting for
                # an image to get pulled vs waiting for a free node in this field, we
                # need to use the Events API to get that information. At some point it
                # may come through in this waiting.reason field, though:
                # https://github.com/kubernetes/kubernetes/issues/19077
                additional_info_builder.append(
                    "(pulling image or waiting for available nodes)"
                )
                is_happy_path = True
        if latest_condition_reason:
            additional_info_builder.append(latest_condition_reason)

        if len(additional_info_builder) == 1:
            additional_info = ""
        else:
            additional_info = " ".join(additional_info_builder)
        if additional_info != prev_additional_info:
            print(f"Waiting for pod {pod_name} to start running{additional_info}")
            prev_additional_info = additional_info
        await asyncio.sleep(1.0)
        i += 1
        if is_happy_path:
            wait_until += 1
        if i > wait_until or i > max_wait_until:
            raise TimeoutError(
                f"Waited >{i} seconds for the container of job {job_id} in pod "
                f"{pod_name} to start running"
            )

        pod = await core_api.read_namespaced_pod_status(pod_name, kubernetes_namespace)
        main_container_state, latest_condition_reason = get_main_container_state(
            pod, job_id, pod_name
        )

    return pod


async def stream_pod_logs(
    core_api: kubernetes_client.CoreV1Api, kubernetes_namespace: str, pod_name: str
) -> None:
    # Now our pod is running, so we can stream the logs

    async with kubernetes_watch.Watch() as w:
        async for line in w.stream(
            core_api.read_namespaced_pod_log,
            name=pod_name,
            namespace=kubernetes_namespace,
        ):
            print(line, end="")


async def wait_for_pod_exit(
    core_api: kubernetes_client.CoreV1Api,
    job_id: str,
    kubernetes_namespace: str,
    pod_name: str,
    timeout_seconds: int,
    streamed_logs: bool,
) -> int:
    # Once this stream ends, we know the pod is completed, but sometimes it takes some
    # time for Kubernetes to report that the pod has completed. So we poll until the pod
    # is reported as terminated.

    pod = await core_api.read_namespaced_pod_status(pod_name, kubernetes_namespace)
    main_container_state, _ = get_main_container_state(pod, job_id, pod_name)
    t0 = time.time()
    while main_container_state is None or main_container_state.running is not None:
        await asyncio.sleep(1.0)
        if time.time() > t0 + timeout_seconds:
            if streamed_logs:
                raise TimeoutError(
                    f"Unexpected. The job {job_id} has a pod {pod_name}, and the pod "
                    f"still seems to be running {timeout_seconds} seconds after the log"
                    " stream ended"
                )
            else:
                raise TimeoutError(
                    f"The job {job_id} timed out, the pod {pod_name} as been running "
                    f"for {timeout_seconds} seconds"
                )
        pod = await core_api.read_namespaced_pod_status(pod_name, kubernetes_namespace)
        main_container_state, _ = get_main_container_state(pod, job_id, pod_name)

    return main_container_state.terminated.exit_code


async def wait_for_pod(
    core_api: kubernetes_client.CoreV1Api,
    job_id: str,
    kubernetes_namespace: str,
    pod: kubernetes_client.V1Pod,
    wait_for_result: WaitOption,
) -> int:
    """
    This function waits for the specified pod to start running, streams the logs from
    that pod into our local stdout, and then waits for the pod to terminate. Then we
    return the exit code of the pod.
    """
    await wait_for_pod_running(core_api, job_id, kubernetes_namespace, pod)
    if wait_for_result == WaitOption.DO_NOT_WAIT:
        # TODO maybe return None instead? Currently this code path is not used, requires
        # support in the caller
        return 0

    if wait_for_result == WaitOption.WAIT_AND_TAIL_STDOUT:
        await stream_pod_logs(core_api, kubernetes_namespace, pod.metadata.name)

        return await wait_for_pod_exit(
            core_api, job_id, kubernetes_namespace, pod.metadata.name, 15, True
        )
    else:
        # TODO this timeout should be configurable and the default should be smaller
        # than 2 days
        wait_for_pod_exit_timeout_seconds = 60 * 60 * 24 * 2
        return await wait_for_pod_exit(
            core_api,
            job_id,
            kubernetes_namespace,
            pod.metadata.name,
            wait_for_pod_exit_timeout_seconds,
            False,
        )


async def get_kubernetes_secret(
    kubernetes_namespace: str, secret_name: str
) -> Dict[str, str]:
    async with kubernetes_client.ApiClient() as api_client:
        core_api = kubernetes_client.CoreV1Api(api_client)
        result = await core_api.read_namespaced_secret(
            secret_name, kubernetes_namespace
        )

    return {
        key: base64.b64decode(value).decode("utf-8")
        for key, value in result.data.items()
    }


async def run_command_on_pod(
    pod_name: str,
    kubernetes_namespace: str,
    command: List[str],
    ws_core_api: kubernetes_client.CoreV1Api,
) -> str:
    """Returns the output of running command"""
    return await ws_core_api.connect_post_namespaced_pod_exec(
        name=pod_name,
        namespace=kubernetes_namespace,
        command=command,
        stderr=True,
        stdin=False,
        stdout=True,
        tty=False,
    )


def get_main_container_is_ready(pod: kubernetes_client.V1Pod) -> bool:
    main_container_statuses = [
        container
        for container in pod.status.container_statuses
        if container.name == "main"
    ]
    if len(main_container_statuses) == 0:
        print(f"Unexpected: pod {pod.metadata.name} has no main container, skipping")
        return False
    elif len(main_container_statuses) > 1:
        print(
            f"Unexpected: pod {pod.metadata.name} has more than one main container, "
            "skipping"
        )
        return False
    else:
        return main_container_statuses[0].ready


async def set_main_container_ready(
    core_api: kubernetes_client.CoreV1Api, pod: kubernetes_client.V1Pod, is_ready: bool
) -> None:
    main_containers = [
        container
        for container in pod.status.container_statuses
        if container.name == "main"
    ]
    if len(main_containers) != 1:
        raise ValueError(
            "Unexpected number of container with the name 'main': "
            f"{len(main_containers)}"
        )

    main_container_state = main_containers[0]
    main_container_state.ready = is_ready

    await core_api.patch_namespaced_pod_status(
        pod.metadata.name,
        pod.metadata.namespace,
        kubernetes_client.V1Pod(
            status=kubernetes_client.V1PodStatus(
                container_statuses=[main_container_state]
            )
        ),
    )



"""
This code belongs in either func_worker_storage.py. Unfortunately, if we put it in
func_worker_storage.py, then meadowrun.__init__ will import k8s_integration.k8s which
will import func_worker_storage, which means that running func_worker_storage will
produce annoying messages like RuntimeWarning: 'meadowrun.func_worker_storage' found in
sys.modules after import of package 'meadowrun', but prior to execution of
'meadowrun.func_worker_storage'; this may result in unpredictable behaviour
warn(RuntimeWarning(msg))
"""
from __future__ import annotations

import dataclasses

from typing import Optional, Tuple, TYPE_CHECKING, Type

from meadowrun.aws_integration.s3 import ensure_uploaded_by_hash
from meadowrun.object_storage import ObjectStorage
from meadowrun.s3_grid_job import read_storage
from meadowrun.storage_keys import STORAGE_CODE_CACHE_PREFIX

if TYPE_CHECKING:
    from types import TracebackType
    import types_aiobotocore_s3

MEADOWRUN_STORAGE_USERNAME = "MEADOWRUN_STORAGE_USERNAME"
MEADOWRUN_STORAGE_PASSWORD = "MEADOWRUN_STORAGE_PASSWORD"

# This is a global variable that will be updated with the storage client if it's
# available in func_worker_storage
FUNC_WORKER_STORAGE_CLIENT: Optional[types_aiobotocore_s3.S3Client] = None
FUNC_WORKER_STORAGE_BUCKET: Optional[str] = None


@dataclasses.dataclass
class FuncWorkerClientObjectStorage(ObjectStorage):
    # this really belongs in k8s.py but can't put it there because of
    # circular imports
    storage_client: Optional[types_aiobotocore_s3.S3Client] = None
    bucket_name: Optional[str] = None

    async def __aexit__(
        self,
        exc_typ: Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        if self.storage_client is not None:
            await self.storage_client.__aexit__(exc_typ, exc_val, exc_tb)

    @classmethod
    def get_url_scheme(cls) -> str:
        return "meadowrunfuncworkerstorage"

    async def _upload(self, file_path: str) -> Tuple[str, str]:
        # TODO these will never get cleaned up

        if self.storage_client is None or self.bucket_name is None:
            raise Exception(
                "Can't use _upload without a bucket_name and a storage endpoint"
            )

        s3_key = await ensure_uploaded_by_hash(
            self.storage_client, file_path, self.bucket_name, STORAGE_CODE_CACHE_PREFIX
        )
        return self.bucket_name, s3_key

    async def _download(
        self, bucket_name: str, object_name: str, file_name: str
    ) -> None:
        storage_client = FUNC_WORKER_STORAGE_CLIENT
        if storage_client is None:
            raise Exception("FUNC_WORKER_STORAGE_CLIENT is not available")
        with open(file_name, "wb") as f:
            f.write(await read_storage(storage_client, bucket_name, object_name))



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['meadowrun',
 'meadowrun._vendor',
 'meadowrun._vendor.aiodocker',
 'meadowrun._vendor.platformdirs',
 'meadowrun.aws_integration',
 'meadowrun.aws_integration.management_lambdas',
 'meadowrun.azure_integration',
 'meadowrun.azure_integration.mgmt_functions',
 'meadowrun.azure_integration.mgmt_functions.azure_core',
 'meadowrun.azure_integration.mgmt_functions.clean_up',
 'meadowrun.azure_integration.mgmt_functions.vm_adjust',
 'meadowrun.deployment',
 'meadowrun.func_worker',
 'meadowrun.k8s_integration']

package_data = \
{'': ['*'], 'meadowrun': ['docker_files/*']}

install_requires = \
['aiobotocore>=2.1.2,<3.0.0',
 'aiohttp>=3.8.0,<4.0.0',
 'asyncssh>=2.11.0,<3.0.0',
 'boto3>=1.21.21,<2.0.0',
 'cloudpickle>=2.0.0,<3.0.0',
 'filelock>=3.6.0,<4.0.0',
 'kubernetes-asyncio>=24.2.0,<25.0.0',
 'protobuf>=3.18.1,<4.0.0',
 'psutil>=5.8.0,<6.0.0',
 'requests>=2.27.1,<3.0.0',
 'typing-extensions>=4.1.1,<5.0.0']

entry_points = \
{'console_scripts': ['meadowrun-local = '
                     'meadowrun.run_job_local_main:command_line_main',
                     'meadowrun-manage-azure-vm = '
                     'meadowrun.manage:main_azure_vm',
                     'meadowrun-manage-ec2 = meadowrun.manage:main_ec2']}

setup_kwargs = {
    'name': 'meadowrun',
    'version': '0.2.10',
    'description': 'The easiest way to run python code on one or more remote machines',
    'long_description': None,
    'author': 'Richard Lee',
    'author_email': 'hrichardlee@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/meadowdata/meadowrun',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
